from datetime import date, datetime
from logging import Logger
from typing import Final
from .datetime_pomes import TIMEZONE_LOCAL
from .env_pomes import APP_PREFIX, env_get_str
from .str_pomes import str_sanitize
from.validation_msgs import _ERR_MSGS_EN, _ERR_MSGS_PT

VALIDATION_MSG_LANGUAGE: Final[str] = env_get_str(f"{APP_PREFIX}_VALIDATION_MSG_LANGUAGE", "pt")
VALIDATION_MSG_PREFIX: Final[str] = env_get_str(f"{APP_PREFIX}_VALIDATION_MSG_PREFIX", APP_PREFIX)


def validate_value(attr: str, val: str | int | float, min_val: int = None,
                   max_val: int = None, default: bool | list[any] = None) -> str:
    """
    Validate *val* according to type, range, or membership in values list, as specified.

    :param attr: the name of the attribute
    :param val: the value to be validated
    :param min_val: if val is a string, specifies its minimum length; otherwise, specifies its minimum value
    :param max_val: if val is a string, specifies its maximum length; otherwise, specifies its maximum value
    :param default: if boolean, requires val to be specified; if list, requires val to be in it
    :return: None if val passes validation, or the corresponding error message otherwise
    """
    # initialize the return variable
    result: str | None = None

    # 'val' can be None, and None can be in 'default'
    if isinstance(default, list):
        if val not in default:
            if val is None:
                result = validate_format_error(105, attr)
            else:
                length: int = len(default)
                # is 'None' the last element in list ?
                if default[-1] is None:
                    # yes, omit it from the message
                    length -= 1
                result = validate_format_error(122, val, attr, [default[:length]])
    elif val is None:
        if isinstance(default, bool) and default:
            result = validate_format_error(105, attr)
    elif isinstance(val, str):
        length: int = len(val)
        if min_val is not None and max_val == min_val and length != min_val:
            result = validate_format_error(109, val, attr, min_val)
        elif max_val is not None and max_val < length:
            result = validate_format_error(108, val, attr, max_val)
        elif min_val is not None and length < min_val:
            result = validate_format_error(107, val, attr, min_val)
    elif (min_val is not None and val < min_val) or \
         (max_val is not None and val > max_val):
        result = validate_format_error(123, val, attr, [min_val, max_val])

    return result


def validate_bool(errors: list[str] | None, scheme: dict, attr: str,
                  default: bool = None, mandatory: bool = False, logger: Logger = None) -> bool:
    """
    Validate the boolean value associated with *attr* in *scheme*.

    If provided, this value must be a *bool*, or the string *t*, *true*, *f*, or *false*.

    :param errors: incidental error messages
    :param scheme: dictionary containing the value to be validated
    :param attr: the name of the attribute whose value is being validated
    :param default: default value, if not found
    :param mandatory: specifies whether the value must be provided
    :param logger: optional logger
    :return: the validated value, or None if validation failed
    """
    # initialize the return variable
    result: bool | None = None

    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]
    try:
        result = scheme[suffix]
        if isinstance(result, str):
            if result.lower() in ["t", "true"]:
                result = True
            elif result.lower() in ["f", "false"]:
                result = False
        if not isinstance(result, bool):
            stat = validate_format_error(124, result, attr, "bool")
    except (KeyError, TypeError):
        if default is not None:
            result = default
        elif mandatory:
            stat = validate_format_error(105, attr)

    if stat:
        __validate_log(errors, stat, logger)

    return result


def validate_int(errors: list[str] | None, scheme: dict, attr: str,
                 min_val: int = None, max_val: int = None,
                 default: bool | int | list[int] = None, logger: Logger = None) -> int:
    """
    Validate the *int* value associated with *attr* in *scheme*.

    If provided, this value must be a *int*, or a valid string representation of a *int*.

    :param errors: incidental error messages
    :param scheme: dictionary containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param min_val: the minimum value accepted
    :param max_val:  the maximum value accepted
    :param default: if int, specifies the default value;
                    if bool, requires the value to be specified;
                    if list, requires the value to be in it
    :param logger: optional logger
    :return: the validated value, or None if validation failed
    """
    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]

    # retrieve the value
    result: int | None = scheme.get(suffix)

    # validate it
    if result is None:
        result = default
    elif isinstance(result, str):
        try:
            result = int(result)
        except ValueError:
            result = None
            stat = validate_format_error(124, result, attr, "int")

    # bool is subtype of int
    if result is not None and \
            (isinstance(result, bool) or not isinstance(result, int)):
        stat = validate_format_error(124, result, attr, "int")

    if not stat:
        stat = validate_value(attr, result, min_val, max_val, default)

    if stat:
        __validate_log(errors, f"{stat} @{attr}", logger)

    return result


def validate_float(errors: list[str] | None, scheme: dict, attr: str,
                   min_val: float = None, max_val: float = None,
                   default: bool | int | float | list[float | int] = None, logger: Logger = None) -> float:
    """
    Validate the *float* value associated with *attr* in *scheme*.

    If provided, this value must be a *float*, or a valid string representation of a *float*.

    :param errors: incidental error messages
    :param scheme: dictionary containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param min_val: the minimum value accepted
    :param max_val:  the maximum value accepted
    :param default: if float, specifies the default value;
                    if bool, requires the value to be specified;
                    if list, requires the value to be in it
    :param logger: optional logger
    :return: the validated value, or None if validation failed
    """
    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]

    # retrieve the value
    result: float | None = scheme.get(suffix)

    # validate it
    if result is None:
        result = default
    elif isinstance(result, str | int):
        try:
            result = float(result)
        except ValueError:
            stat = validate_format_error(124, result, attr, "int")

    if result is not None and not isinstance(result, float):
        stat = validate_format_error(124, result, attr, "float")

    if not stat:
        stat = validate_value(attr, result, min_val, max_val, default)

    if stat:
        result = None
        __validate_log(errors, stat, logger)

    return result


def validate_str(errors: list[str] | None, scheme: dict, attr: str,
                 min_length: int = None, max_length: int = None,
                 default: bool | str | list[str] = None, logger: Logger = None) -> str:
    """
    Validate the *str* value associated with *attr* in *scheme*.

    If provided, this value must be a *str*.

    :param errors: incidental error messages
    :param scheme: dictionary containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param min_length: the minimum length accepted
    :param max_length:  the maximum length accepted
    :param default: if str, specifies the default value;
                    if bool, requires the value to be specified;
                    if list, requires the value to be in it
    :param logger: optional logger
    :return: the validated value, or None if validation failed
    """
    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]

    # obtain and validate the value
    result: str = scheme.get(suffix)
    if result and not isinstance(result, str):
        stat = validate_format_error(124, result, attr, "str")
    elif isinstance(default, str):
        if result is None:
            result = default
        else:
            stat = validate_value(attr, result, min_length, max_length)
    else:
        stat = validate_value(attr, result, min_length, max_length, default)

    if stat:
        __validate_log(errors, stat, logger)

    return result


def validate_date(errors: list[str] | None, scheme: dict, attr: str,
                  day_first: bool = False, default: bool | date = None, logger: Logger = None) -> date:
    """
    Validate the *date* value associated with *attr* in *scheme*.

    If provided, this value must be a *date*, or a valid string representation of a *date*.

    :param errors: incidental error messages
    :param scheme: dictionary containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param day_first: indicates that the day precedes the month in the string representing the date
    :param default: if date, specifies the default value;
                    if bool, requires the value to be specified
    :param logger: optional logger
    :return: the validated value, or None if validation failed
    """
    # import needed module
    from .datetime_pomes import date_parse

    # initialize the return variable
    result: date | None = None

    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]
    try:
        date_str: str = scheme[suffix]
        result = date_parse(date_str, dayfirst=day_first)
        if result is None:
            stat = validate_format_error(106, date_str, attr)
        elif result > datetime.now(TIMEZONE_LOCAL).date():
            stat = validate_format_error(110, date_str, attr)
    except KeyError:
        if isinstance(default, bool) and default:
            stat = validate_format_error(105, attr)
        elif isinstance(default, date):
            result = default

    if stat:
        __validate_log(errors, stat, logger)

    return result


def validate_datetime(errors: list[str] | None, scheme: dict, attr: str,
                      day_first: bool = True, default: bool | datetime = None, logger: Logger = None) -> datetime:
    """
    Validate the *datetime* value associated with *attr* in *scheme*.

    If provided, this value must be a *date*, or a valid string representation of a *date*.

    :param errors: incidental error messages
    :param scheme: dictionary containing the value to be validated
    :param attr: the attribute associated with the value to be validated
    :param day_first: indicates that the day precedes the month in the string representing the date
    :param default: if datetime, specifies the default value;
                    if bool, requires the value to be specified
    :param logger: optional logger
    :return: the validated value, or None if validation failed
    """
    # import needed module
    from .datetime_pomes import datetime_parse

    # initialize the return variable
    result: datetime | None = None

    stat: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]
    try:
        date_str: str = scheme[suffix]
        result = datetime_parse(date_str, dayfirst=day_first)
        if result is None:
            stat = validate_format_error(106, date_str, attr)
        elif result > datetime.now(TIMEZONE_LOCAL):
            stat = validate_format_error(110, date_str, attr)
    except KeyError:
        if isinstance(default, bool) and default:
            stat = validate_format_error(105, attr)
        elif isinstance(default, datetime):
            result = default

    if stat:
        __validate_log(errors, stat, logger)

    return result


def validate_ints(errors: list[str] | None, scheme: dict, attr: str,
                  min_val: int = None, max_val: int = None,
                  mandatory: bool = False, logger: Logger = None) -> list[int]:
    """
    Validate the list of *int* values associated with *attr* in *scheme*.

    If provided, this list must contain *ints*, or valid string representations of *ints*.

    :param errors: incidental error messages
    :param scheme: dictionary containing the list of values to be validated
    :param attr: the attribute associated with the list of values to be validated
    :param min_val: the minimum value accepted
    :param max_val:  the maximum value accepted
    :param mandatory: whether the list of values must be provided
    :param logger: optional logger
    :return: the list of validated values, or None if validation failed
    """
    # initialize the return variable
    result: list[any] | None = None

    err_msg: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]
    try:
        values: list[any] = scheme[suffix]
        if isinstance(values, list):
            result = []
            if len(values) > 0:
                for inx, value in enumerate(values):
                    result.append(value)
                    if isinstance(value, int):
                        err_msg = validate_value(f"@{attr}[{inx+1}]", value, min_val, max_val)
                    else:
                        err_msg = validate_format_error(124, value, f"@{attr}[{inx+1}]", "int")
            elif mandatory:
                err_msg = validate_format_error(105, attr)
        else:
            err_msg = validate_format_error(124, result, attr, "list")
    except (KeyError, TypeError):
        if mandatory:
            err_msg = validate_format_error(105, attr)

    if err_msg:
        __validate_log(errors, err_msg, logger)

    return result


def validate_strs(errors: list[str] | None, scheme: dict,
                  attr: str, min_length: int, max_length: int,
                  mandatory: bool = False, logger: Logger = None) -> list[str]:
    """
    Validate the list of *str* values associated with *attr* in *scheme*.

    If provided, this list must contain *strs*.

    :param errors: incidental error messages
    :param scheme: dictionary containing the list of values to be validated
    :param attr: the attribute associated with the list of values to be validated
    :param min_length: the minimum length accepted
    :param max_length:  the maximum length accepted
    :param mandatory: whether the list of values must be provided
    :param logger: optional logger
    :return: the list of validated values, or None if validation failed
    """
    # initialize the return variable
    result: list[any] | None = None

    err_msg: str | None = None
    pos: int = attr.rfind(".") + 1
    suffix: str = attr[pos:]
    try:
        values: list[any] = scheme[suffix]
        if isinstance(values, list):
            result = []
            if len(values) > 0:
                for inx, value in enumerate(values):
                    result.append(value)
                    if isinstance(value, str):
                        err_msg = validate_value(f"@{attr}[{inx+1}]", value, min_length, max_length)
                    else:
                        err_msg = validate_format_error(124, value, f"@{attr}[{inx+1}]", "str")
            elif mandatory:
                err_msg = validate_format_error(105, attr)
        else:
            err_msg = validate_format_error(124, result, attr, "list")
    except (KeyError, TypeError):
        if mandatory:
            err_msg = validate_format_error(105, attr)

    if err_msg:
        __validate_log(errors, err_msg, logger)

    return result


def validate_format_error(error_id: int, *args) -> str:
    """
    Format and return the error message identified by *err_id* in the standard messages list.

    The message is built from the message element in the standard messages list, identified by *err_id*.
    The occurrences of '{}' in the element are sequentially replaced by the given *args*.

    :param error_id: the identification of the message element
    :param args: optional arguments to format the error message with
    :return: the formatted error message
    """
    # retrieve the standard validation messages list
    match VALIDATION_MSG_LANGUAGE:
        case "en":
            err_msgs = _ERR_MSGS_EN
        case "pt":
            err_msgs = _ERR_MSGS_PT
        case _:
            err_msgs = {}

    # initialize the return variable
    result: str = VALIDATION_MSG_PREFIX + str(error_id) + ": " + err_msgs.get(error_id)

    # apply the provided arguments
    for arg in args:
        if arg is None:
            result = result.replace(" {}", "", 1)
        elif isinstance(arg, str) and arg.find(" ") > 0:
            result = result.replace("{}", arg, 1)
        else:
            result = result.replace("{}", f"'{arg}'", 1)

    return result


def validate_format_errors(errors: list[str]) -> list[dict]:
    """
    Build and return a list of dicts from the list of errors *errors*.

    Each element in *errors* is encoded as a *dict*.
    This list is tipically used in a returning *JSON* string.

    :param errors: the list of errors to build the list of dicts with
    :return: the built list
    """
    # initialize the return variable
    result: list[dict] = []

    # extract error code, description, and attribute from text
    for error in errors:
        desc: str = error
        out_error: dict = {}

        # does the text contain an error code ?
        if desc.startswith(VALIDATION_MSG_PREFIX):
            # yes
            term: str = "code" if VALIDATION_MSG_LANGUAGE == "en" else "codigo"
            pos: int = desc.find(":")
            out_error[term] = desc[0:pos]
            desc = desc[pos+2:]

        term: str = "description" if VALIDATION_MSG_LANGUAGE == "en" else "descricao"
        out_error[term] = desc
        result.append(out_error)

    return result


def validate_unformat_errors(errors: list[dict | str]) -> list[str]:
    """
    Extract and return the list of errors used to build the list of dicts *errors*.

    :param errors: the list of dicts to extract the errors from
    :return: the built list
    """
    # initialize the return variable
    result: list[str] = []

    # define the dictionary keys
    name: str = "code" if VALIDATION_MSG_LANGUAGE == "en" else "codigo"
    desc: str = "description" if VALIDATION_MSG_LANGUAGE == "en" else "descricao"

    # traverse the list of dicts
    for error in errors:
        if isinstance(error, dict):
            result.append(f"{error.get(name)}: {str_sanitize(error.get(desc))}")
        else:
            result.append(error)

    return result


def __validate_log(errors: list[str], err_msg: str, logger: Logger) -> None:
    """
    LOg *err_msg* using *logger*, and add it to the error messages list *errors*.

    :param errors: the error messages list
    :param err_msg: the message to log
    :param logger: the logger
    """
    if logger:
        logger.error(err_msg)
    if errors is not None:
        errors.append(err_msg)



import pytz
from datetime import date, datetime
from dateutil import parser
from typing import Final

from .env_pomes import APP_PREFIX, env_get_str

# some useful date/datetime formats
DATE_FORMAT_STD: Final[str] = "%m/%d/%Y"
DATE_FORMAT_COMPACT: Final[str] = "%Y%m%d"
DATE_FORMAT_INV: Final[str] = "%Y-%m-%d"
DATETIME_FORMAT_STD: Final[str] = "%m/%d/%Y %H:%M:%S"
DATETIME_FORMAT_COMPACT: Final[str] = "%Y%m%d%H%M%S"
DATETIME_FORMAT_INV: Final[str] = "%Y-%m-%d %H:%M:%S"

TIMEZONE_LOCAL: pytz.BaseTzInfo = pytz.timezone(env_get_str(f"{APP_PREFIX}_TIMEZONE_LOCAL", "America/Sao_Paulo"))
TIMEZONE_UTC: pytz.BaseTzInfo = pytz.UTC


def date_reformat(dt_str: str, to_format: str, **kwargs: any) -> str:
    """
    Convert the date in *dt_str* to the format especified in *to_format*.

    The argument *dt_str* must represent a valid date, with or without time-of-day information.
    The argument *to_format* must be a valid format for *date* ou *datetime*.

    In *kwargs*, it may optionally be specified:
        -   *dayfirst*, to signal that *day* comes before *month* in *dt_str*
        -   *fmt=<format>*, to force use of a specific format

    :param dt_str: the date to convert
    :param to_format: the format for the conversion
    :param kwargs: optional arguments for the parser in python-dateutil
    :return:  the converted date
    """
    result: str | None = None
    ts: datetime = parser.parse(dt_str, **kwargs)
    if ts:
        result = datetime.strftime(ts, to_format)

    return result


def date_parse(dt_str: str, **kwargs: any) -> date:
    """
    Obtain and return the *date* object corresponding to *dt_str*.

    In *kwargs*, it may optionally be specified:
        -   *dayfirst*, to signal that *day* comes before *month* in *dt_str*
        -   *fmt=<format>*, to force use of a specific format
    Return *None* se *dt_str* does not contain a valid date.

    :param dt_str: the date, in a supported format
    :param kwargs: optional arguments for the parser in python-dateutil
    :return: the corresponding date object, or None
    """
    result: date | None = None
    if dt_str:
        result = parser.parse(dt_str, **kwargs).date()

    return result


def datetime_parse(dt_str: str, **kwargs: any) -> datetime:
    """
    Obtain and return the *datetime* object corresponding to *dt_str*.

    In *kwargs*, it may optionally be specified:
        -   *dayfirst*, to signal that *day* comes before *month* in *dt_str*
        -   *fmt=<format>*, to force use of a specific format
   Return *None* se *dt_str* does not contain a valid date.

    :param dt_str: the date, in a supported format
    :param kwargs: optional arguments for the parser in python-dateutil
    :return: the corresponding datetime object, or None
    """
    result: datetime | None = None
    if dt_str:
        result = parser.parse(dt_str, **kwargs)

    return result



def list_compare(list1: list[any], list2: list[any]) -> bool:
    """
    Compare the contents of the two lists *list1* e *list2*.

    Return *True* if all the elements in *list1* are also in *list2*, and vice-versa, with the same cardinality.

    :param list1: the first list
    :param list2: the second list
    :return: True if the two lists contain the same elements, in the same quantity, in any order
    """
    # initialize the return variable
    result: bool = True

    # are the input parameters lists containing the same number of elements ?
    if isinstance(list1, list) and isinstance(list2, list) and len(list1) == len(list2):
        # yes, verify whether all elements in 'list1' are also in 'list2', in the same quantity
        for elem in list1:
            # is 'elem' in both lists, in thew same quantity ?
            if list1.count(elem) != list2.count(elem):
                # no, the lists are not equal
                result = False
                break
    else:
        # no, the lists are not equal
        result = False

    return result


def list_flatten(source: list[str]) -> str:
    """
    Transforma uma lista de *str* em uma *str* consistindo na concatenação com "." dos elementos da lista.

    Exemplos:
        - ['1', '2', '']     -> '1.2.'
        - ['', 'a', 'b']     -> '.a.b'
        - ['x', '', '', 'y'] -> 'x...y'
        - ['z']              -> 'z'

    :param source: a lista de strings
    :return: a string concatenada
    """
    result: str = ""
    for item in source:
        result += "." + item
    return result[1:]


def list_unflatten(source: str) -> list[str]:
    """
    Transforma uma *str* contendo elementos concatenados por "." em uma lista de *str*.

    Essa lista contem os sub_elementos extraídos. Exemplos:
        - '1.2.'  -> ['1', '2', '']
        - '.a.b'  -> ['', 'a', 'b']
        - 'x...y' -> ['x', '', '', 'y']
        - 'z'     -> ['z']

    :param source: string com elementos concatenados por "."
    :return: a lista de strings contendo os elementos da concatenação
    """
    # import the needed function
    from .str_pomes import str_split_on_mark

    return str_split_on_mark(source, ".")


def list_find_coupled(coupled_elements: list[tuple[str, str]], primary_element: str) -> str:
    """
    Localiza em *coupled_elements* o elemento acoplado ao *primary_element* dado.

    Se *primary_element* contiver indicação de índice (denotado por *[<pos>]*), essa indicação é removida.
    Essa função é utilizada na transformação de *dicts* (*dict_transform*) e *lists* (*list_transform*),
    a partir de sequências de pares de chaves.

    :param coupled_elements: a lista de tuplas contendo os pares de elementos.
    :param primary_element: o elemento primário
    :return: o elemento acoplado, ou None se não for encontrado
    """
    # remove a indicação de elemento de lista
    pos1: int = primary_element.find("[")
    while pos1 > 0:
        pos2: int = primary_element.find("]", pos1)
        primary_element = primary_element[:pos1] + primary_element[pos2+1:]
        pos1 = primary_element.find("[")

    # inicializa a variável de retorno
    result: str | None = None

    # percorre a list de elementos acoplados
    for primary, coupled in coupled_elements:
        # o elemento primário foi encontrado ?
        if primary == primary_element:
            # sim, retorne o elemento acoplado correspondente
            result = coupled
            break

    return result


def list_transform(source: list[any], from_to_keys: list[tuple[str, str]],
                   prefix_from: str = None, prefix_to: str = None) -> list[any]:
    """
    Constrói uma nova *list*, transformando elementos do tipo *list* e *dict* encontrados em *source*.

    A conversão dos elementos tipo *dict* está documentada na função *dict_transform*.

    Os prefixos para as chaves de origem e de destino, se definidos, tem tratamentos distintos.
    São acrescentados na busca de valores em *Source*, e removidos na atribuição de valores
    ao *dict* de retorno.

    :param source: o dict de origem dos valores
    :param from_to_keys: a lista de tuplas contendo as sequências de chaves de origem e destino
    :param prefix_from: prefixo a ser acrescentado às chaves de origem
    :param prefix_to: prefixo a ser removido das chaves de destino
    :return: a nova lista
    """
    # import the needed function
    from .dict_pomes import dict_transform

    # inicializa a variável de retorno
    result: list[any] = []

    # percorre a lista de origem
    for inx, value in enumerate(source):
        if prefix_from is None:
            from_keys: None = None
        else:
            from_keys: str = f"{prefix_from}[{inx}]"

        # obtem o valor de destino
        if isinstance(value, dict):
            to_value: dict = dict_transform(value, from_to_keys, from_keys, prefix_to)
        elif isinstance(value, list):
            to_value: list = list_transform(value, from_to_keys, from_keys, prefix_to)
        else:
            to_value: any = value

        # acrescenta o valor transformado ao resultado
        result.append(to_value)

    return result


def list_elem_starting_with(source: list[str | bytes],
                            prefix: str | bytes, keep_prefix: bool = True) -> str | bytes:
    """
    Localiza e retorna o primeiro elemento em *source* prefixado por *prefix*.

    Retorna *None* se esse elemento não for encontrado.

    :param source: a lista a ser inspecionada
    :param prefix: o dado prefixando o elemento a ser retornado
    :param keep_prefix: define se o elemento encontrado deve ou não ser retornado com o prefixo
    :return: o elemento prefixado, com ou sem o prefixo
    """
    # inicializa a variável de retorno
    result: str | bytes | None = None

    # percorre a lista de origem
    for elem in source:
        if elem.startswith(prefix):
            if keep_prefix:
                result = elem
            else:
                result = elem[len(prefix)+1:]
            break

    return result



from types import TracebackType
import os
import traceback


def exc_format(exc: Exception,
               exc_info: tuple[type[BaseException], BaseException, TracebackType]) -> str:
    """
    Format the error message resulting from the exception raised in execution time.

    The format to use: <python_module>, <line_number>: <exc_class> - <exc_text>

    :param exc: the exception raised
    :param exc_info: information associated with the exception
    :return: the formatted message
    """
    tback: TracebackType = exc_info[2]
    cls: str = str(exc.__class__)

    # retrieve the execution point where the exception was raised (bottom of the stack)
    tlast: traceback = tback
    while tlast.tb_next:
        tlast = tlast.tb_next

    # retrieve the module name and the line number within the module
    try:
        fname: str = os.path.split(tlast.tb_frame.f_code.co_filename)[1]
    except Exception:
        fname: str = "<unknow module>"
    fline: int = tlast.tb_lineno
    return f"{fname}, {fline}, {cls[8:-2]} - {exc}"



from pathlib import Path
from typing import Final
from xmltodict3 import XmlTextToDict

XML_FILE_HEADER: Final[str] = '<?xml version="1.0" encoding="UTF-8" ?>'


def xml_normalize_keys(source: dict) -> dict:
    """
    Clona o *source*, removenda os *namespaces* e os prefixos *'@'* e *'#'* nos nomes de suas chaves.

    A ordem das chaves é mantida.

    :param source: o dict de referência
    :return: um novo dict normalizado
    """
    # inicializa a variável de retorno
    result: dict = {}

    # percorre o dicionário
    for curr_key, curr_value in source.items():

        # o valor atual é um dicionário ?
        if isinstance(curr_value, dict):
            # sim, prossiga recursivamente
            result[curr_key] = xml_normalize_keys(curr_value)
        # o valor atual é uma lista ?
        elif isinstance(curr_value, list):
            # sim, percorra-a
            result[curr_key] = []
            for item in curr_value:
                # o item da lista é um dicionário ?
                if isinstance(item, dict):
                    # sim, prossiga recursivamente
                    result[curr_key].append(xml_normalize_keys(item))
                else:
                    result[curr_key].append(item)
        # ä chave atual tem prefixo a ser removido ?
        elif curr_key[0:1] in ["@", "#"]:
            # sim
            result[curr_key[1:]] = curr_value
        else:
            pos: int = curr_key.find(":")
            if pos == 0:
                result[curr_key] = curr_value
            else:
                result[curr_key[pos+1:]] = curr_value

    return result


def xml_to_dict(file_data: bytes | str) -> dict:
    """
    Convert the XML into a *dict*, by removing namespaces, and keys prefixed with "@" e "#".

    O XML de entrada deve estar em *file_data* (tipo *bytes*),
    ou em arquivo do sistema com o caminho especificado por *file_data* (tipo *str*).

    :param file_data: XML a ser convertido
    :return: dict normalizado
    """
    # file_data é o próprio conteúdo XML ?
    if isinstance(file_data, bytes):
        # sim
        file_bytes: bytes = file_data
    else:  # elif isinstance(file_date, str):
        # não, obtenha-o do arquivo
        with Path.open(Path(file_data), "rb") as f:
            file_bytes: bytes = f.read()

    # converte o XML em dict
    xml_data = XmlTextToDict(xml_text=file_bytes.decode(),
                             ignore_namespace=True)
    result: dict = xml_data.get_dict()

    # normaliza o dict, removendo namespaces e prefixos "@" e "#" nos nomes das chaves
    return xml_normalize_keys(result)



from typing import Final

__ERR_MSGS: Final[dict] = {
    101: {
        "en": "Error accessing the DB {} in {}: {}",
        "pt": "Erro na interação com o BD {} em {}: {}",
    },
    102: {
        "en": "Error accessing the object store: {}",
        "pt": "Erro na interação com o armazenador de objetos: {}",
    },
    103: {
        "en": "No {} found",
        "pt": "Nenhum(a) {} encontrado(a)",
    },
    104: {
        "en": "Unknown parameter: {}",
        "pt": "Parâmetro desconhecido: {}",
    },
    105: {
        "en": "Mandatory parameter: {}",
        "pt": "Parâmetro obrigatório: {}",
    },
    106: {
        "en": "Invalid value {} for parameter {}",
        "pt": "Valor {} inválido para o parâmetro {}",
    },
    107: {
        "en": "Invalid value {}: length shorter than {}",
        "pt": "Valor {} inválido: comprimento menor que {}",
    },
    108: {
        "en": "Invalid value {} for parameter {}: length longer than {}",
        "pt": "Valor {} inválido para o parâmetro {}: comprimento maior que {}",
    },
    109: {
        "en": "Invalid value {} for parameter {}: length must be {}",
        "pt": "Valor {} inválido para o parâmetro {}: comprimento deve ser igual a {}",
    },
    110: {
        "en": "Invalid value {} for parameter {}: date is later than the current date",
        "pt": "Valor {} inválido para o parâmetro {}: data posterior à data atual",
    },
    111: {
        "en": "Invalid value {} for parameter {}: wrong date format",
        "pt": "Valor {} inválido para o parâmetro {}: formato de data inválido",
    },
    112: {
        "en": "Error executing operation {}: {}",
        "pt": "Erro ao executar a operação {}: {}",
    },
    113: {
        "en": "The operation {} returned the error {}",
        "pt": "A operação {} retornou o erro {}",
    },
    114: {
        "en": "Unexpected error: {}",
        "pt": "Erro não previsto: {}",
    },
    115: {
        "en": "No file found at URL {}",
        "pt": "Nenhum arquivo encontrado na URL {}",
    },
    116: {
        "en": "The format for file {} is not type {}",
        "pt": "O formato do arquivo {} não é do tipo {}",
    },
    117: {
        "en": "Register for {} already exists",
        "pt": "Registro de {} já existe",
    },
    118: {
        "en": "Authentication token required",
        "pt": "Token de autenticação requerido",
    },
    119: {
        "en": "Invalid authentication token",
        "pt": "Token de autenticação inválido",
    },
    120: {
        "en": "More than one {} found",
        "pt": "Mais de um(a) {} encontrado(a)",
    },
    121: {
        "en": "Invalid operation: {}",
        "pt": "Operação inválida: {}",
    },
    122: {
        "en": "Invalid value {} for parameter {}: must be one of {}",
        "pt": "Valor {} inválido para o parâmetro {}: deve ser um de {}",
    },
    123: {
        "en": "Invalid value {} for parameter {}: must be in the range {}",
        "pt": "Valor {} inválido para o parâmetro {}: deve estar no intervalo {}",
    },
    124: {
        "en": "Invalid value {} for parameter {}: must be type {}",
        "pt": "Valor {} inválido para o parâmetro {}: deve ser do tipo {}",
    },
    125: {
        "en": "The values provided {} for parameter {} do not constitute a valid set of values",
        "pt": "Os valores fornecidos {} para o parâmetro {} não formam um conjunto válido de valores",
    },
    140: {
        "en": "Error retrieving environment variable {}: {}",
        "pt": "Erro na recuperação da variável de ambiente {}: {}",
    },
    141: {
        "en": "Error invoking service {}: {}",
        "pt": "Erro na invocação do serviço {}: {}",
    },
}
_ERR_MSGS_EN: dict = {}
for key, value in __ERR_MSGS.items():
    _ERR_MSGS_EN[key] = value["en"]

_ERR_MSGS_PT: dict = {}
for key, value in __ERR_MSGS.items():
    _ERR_MSGS_PT[key] = value["pt"]


def validation_add_msgs(msgs: dict, lang: str = "en") -> None:
    """
    Add the messages in *msgs* to the standard validation messages list for language *lang".

    :param msgs: list of messages to be added
    :param lang: the reference language
    """
    match lang:
        case "en":
            _ERR_MSGS_EN.update(msgs)
        case "pt":
            _ERR_MSGS_PT.update(msgs)


def validation_set_msgs(msgs: dict, lang: str = "en") -> None:
    """
    Set  the standard validation messages list for language *lang* to the messages in *msgs*.

    :param msgs: list of messages to set the standard validation messages to
    :param lang: the reference language
    """
    global _ERR_MSGS_EN, _ERR_MSGS_PT

    match lang:
        case "en":
            _ERR_MSGS_EN = msgs
        case "pt":
            _ERR_MSGS_PT = msgs



import os
from pathlib import Path


def env_get_str(key: str, def_value: str = None) -> str:
    """
    Retrieve and return the string value defined for *key* in the current operating environment.

    :param key: The key the value is associated with
    :param def_value: The default value to return, if the key has not been defined
    :return: The str value associated with the key
    """
    result: str
    try:
        result = os.environ[key]
    except (KeyError, TypeError):
        result = def_value

    return result


def env_get_bool(key: str, def_value: bool = None) -> bool:
    """
    Retrieve and return the boolean value defined for *key* in the current operating environment.

    :param key: The key the value is associated with
    :param def_value: The default value to return, if the key has not been defined
    :return: The bool value associated with the key
    """
    result: bool
    try:
        result = bool(int(os.environ[key]))
    except (KeyError, TypeError):
        result = def_value

    return result


def env_get_int(key: str, def_value: int = None) -> int:
    """
    Retrieve and return the int value defined for *key* in the current operating environment.

    :param key: The key the value is associated with
    :param def_value: The default value to return, if the key has not been defined
    :return: The int value associated with the key
    """
    result: int
    try:
        result = int(os.environ[key])
    except (KeyError, TypeError):
        result = def_value

    return result


def env_get_float(key: str, def_value: float = None) -> float:
    """
    Retrieve and return the float value defined for *key* in the current operating environment.

    :param key: The key the value is associated with
    :param def_value: The default value to return, if the key has not been defined
    :return: The float value associated with the key
    """
    result: float
    try:
        result = int(os.environ[key])
    except (KeyError, TypeError):
        result = def_value

    return result


def env_get_path(key: str, def_value: Path = None) -> Path:
    """
    Retrieve and return the path value defined for *key* in the current operating environment.

    :param key: The key the value is associated with
    :param def_value: The default value to return, if the key has not been defined
    :return: The path value associated with the key
    """
    result: Path
    try:
        result = Path(os.environ[key])
    except (KeyError, TypeError):
        result = def_value

    return result


# the prefix to the names of the environment variables
APP_PREFIX = env_get_str("PYPOMES_APP_PREFIX", "PYPOMES")



import base64
from collections.abc import Iterable


def json_normalize_dict(source: dict) -> None:
    """
    Turn the values in *source* into values that can be serialized to JSON, thus avoiding *TypeError*.

    Possible transformations:
        - *bytes* e *bytearray* are changed to *str* in *Base64* format
        - *Iterable* is changed into a *list*
        - all other types are left unchanged
    HAZARD: depending on the type of object contained in *source*, the final result may not be serializable.

    :param source: the dict to be made serializable
    """
    for key, value in source.items():
        if isinstance(value, dict):
            json_normalize_dict(value)
        elif isinstance(value, bytes | bytearray):
            source[key] = base64.b64encode(value).decode()
        elif isinstance(value, Iterable) and not isinstance(value, str):
            source[key] = json_normalize_iterable(value)


def json_normalize_iterable(source: Iterable) -> list[any]:
    """
    Return in a *list* the values in *source* that can be serialized to JSON, thus avoiding *TypeError*.

    Possible operations:
        - *bytes* e *bytearray* are changed to *str* in *Base64* format
        - *Iterable* is changed into a *list*
        - all other types are left unchanged
    HAZARD: depending on the type of object contained in *source*, the final result may not be serializable.

    :param source: the dict to be made serializable
    :return: list with serialized values
    """
    result: list[any] = []
    for value in source:
        if isinstance(value, dict):
            json_normalize_dict(value)
            result.append(value)
        elif isinstance(value, bytes | bytearray):
            result.append(base64.b64encode(value).decode())
        elif isinstance(value, Iterable) and not isinstance(value, str):
            result.append(json_normalize_iterable(value))
        else:
            result.append(value)

    return result



from .datetime_pomes import (
    DATE_FORMAT_STD, DATE_FORMAT_COMPACT, DATE_FORMAT_INV,
    DATETIME_FORMAT_STD, DATETIME_FORMAT_COMPACT, DATETIME_FORMAT_INV,
    TIMEZONE_LOCAL, TIMEZONE_UTC,
    date_reformat, date_parse, datetime_parse,
)
from .dict_pomes import (
    dict_has_key_chain, dict_get_value, dict_set_value, dict_reduce,
    dict_listify, dict_transform, dict_merge, dict_coalesce, dict_clone,
    dict_get_key, dict_get_keys, dict_from_object, dict_from_list, dict_replace_value, dict_pop_value,
)
from .email_pomes import (
    EMAIL_ACCOUNT, EMAIL_PWD, EMAIL_PORT, EMAIL_SERVER, email_send,
)
from .encoding_pomes import (
    encode_ascii_hex, decode_ascii_hex,
)
from .env_pomes import (
    APP_PREFIX, env_get_str, env_get_int, env_get_bool, env_get_float, env_get_path,
)
from .exception_pomes import (
    exc_format,
)
from .file_pomes import (
    TEMP_DIR,
    file_from_request, file_get_data
)
from .json_pomes import (
    json_normalize_dict, json_normalize_iterable,
)
from .logging_pomes import (
    LOGGING_ID, LOGGING_LEVEL, LOGGING_FORMAT, LOGGING_STYLE,
    LOGGING_FILE_PATH, LOGGING_FILE_MODE, PYPOMES_LOGGER,
    logging_get_entries, logging_get_entries_from_request,
    logging_log_msgs, logging_log_debug, logging_log_error,
    logging_log_info, logging_log_critical, logging_log_warning,
)
from .list_pomes import (
    list_compare, list_flatten, list_unflatten,
    list_find_coupled, list_elem_starting_with, list_transform,
)
from .str_pomes import (
    str_sanitize, str_between, str_split_on_mark, str_find_whitespace,
)
from .validation_msgs import (
    validation_add_msgs, validation_set_msgs
)
from .validation_pomes import (
    VALIDATION_MSG_LANGUAGE, VALIDATION_MSG_PREFIX,
    validate_value, validate_bool, validate_int, validate_float, validate_str,
    validate_date, validate_datetime, validate_ints, validate_strs,
    validate_format_error, validate_format_errors, validate_unformat_errors,
)
from .xml_pomes import (
    XML_FILE_HEADER, xml_to_dict, xml_normalize_keys,
)

__all__ = [
    # datetime_pomes
    "DATE_FORMAT_STD", "DATE_FORMAT_COMPACT", "DATE_FORMAT_INV",
    "DATETIME_FORMAT_STD", "DATETIME_FORMAT_COMPACT", "DATETIME_FORMAT_INV",
    "TIMEZONE_LOCAL", "TIMEZONE_UTC",
    "date_reformat", "date_parse", "datetime_parse",
    # dict_pomes
    "dict_has_key_chain", "dict_get_value", "dict_set_value", "dict_reduce",
    "dict_listify", "dict_transform", "dict_merge", "dict_coalesce", "dict_clone",
    "dict_get_key", "dict_get_keys", "dict_from_object", "dict_from_list",
    "dict_replace_value", "dict_pop_value",
    # email_pomes
    "EMAIL_ACCOUNT", "EMAIL_PWD", "EMAIL_PORT", "EMAIL_SERVER", "email_send",
    # encoding_pomes
    "encode_ascii_hex", "decode_ascii_hex",
    # env_pomes
    "APP_PREFIX", "env_get_str", "env_get_int", "env_get_bool", "env_get_float", "env_get_path",
    # exception_pomes
    "exc_format",
    # file_pomes
    "TEMP_DIR",
    "file_from_request", "file_get_data",
    # json_pomes
    "json_normalize_dict", "json_normalize_iterable",
    # logging_pomes
    "LOGGING_ID", "LOGGING_LEVEL", "LOGGING_FORMAT", "LOGGING_STYLE",
    "LOGGING_FILE_PATH", "LOGGING_FILE_MODE", "PYPOMES_LOGGER",
    "logging_get_entries", "logging_get_entries_from_request",
    "logging_log_msgs", "logging_log_debug", "logging_log_error",
    "logging_log_info", "logging_log_critical", "logging_log_warning",
    # list_pomes
    "list_compare", "list_flatten", "list_unflatten",
    "list_find_coupled", "list_elem_starting_with", "list_transform",
    # str_pomes
    "str_sanitize", "str_between", "str_split_on_mark", "str_find_whitespace",
    # validation_msgs
    "validation_add_msgs", "validation_set_msgs",
    # validation_pomes
    "VALIDATION_MSG_LANGUAGE", "VALIDATION_MSG_PREFIX",
    "validate_value", "validate_bool", "validate_int", "validate_float", "validate_str",
    "validate_date", "validate_datetime", "validate_ints", "validate_strs",
    "validate_format_error", "validate_format_errors", "validate_unformat_errors",
    # xml_pomes
    "XML_FILE_HEADER", "xml_to_dict", "xml_normalize_keys",
]

from importlib.metadata import version
__version__ = version("pypomes_core")
__version_info__ = tuple(int(i) for i in __version__.split(".") if i.isdigit())



def str_sanitize(target_str: str) -> str:
    """
    Clean the given *target_str* string.

    The sanitization is carried out by:
        - removing backslashes
        - replacing double quotes with single quotes
        - replacing newlines and tabs with whitespace
        - replacing multiple consecutive spaces with a single space

    :param target_str: the string to be cleaned
    :return: the cleaned string
    """
    cleaned: str = target_str.replace("\\", "") \
                             .replace('"', "'") \
                             .replace("\n", " ") \
                             .replace("\t", " ")
    return " ".join(cleaned.split())


def str_split_on_mark(source: str, mark: str) -> list[str]:
    """
    Extrai, de *source*, os segmentos de texto separados por *mark*, e os retorna em uma lista.

    Os segmentos retornados não contem o separador.

    :param source: o texto de referência
    :param mark: o separador
    :return: a lista de segmentos de texto obtidos
    """
    # inicializa a variável de retorno
    result: list[str] = []

    pos: int = 0
    skip: int = len(mark)
    after: int = source.find(mark)
    while after >= 0:
        result.append(source[pos:after])
        pos = after + skip
        after = source.find(mark, pos)
    if pos < len(source):
        result.append(source[pos:])
    else:
        result.append("")

    return result


def str_between(source: str, from_str: str, to_str: str) -> str:
    """
    Extrai e retorna a *substring* em *source* localizada entre os delimitadores *from_st* e *to_str*.

    Retorna *None* se essa extração não for possível.

    :param source: a string a ser pesquisada
    :param from_str: o delimitador inicial
    :param to_str: o delimitador final
    :return: a substring procurada
    """
    # inicializa a variável de retorno
    result: str | None = None

    pos1: int = source.find(from_str)
    if pos1 >= 0:
        pos1 += len(from_str)
        pos2: int = source.find(to_str, pos1)
        if pos2 >= pos1:
            result = source[pos1:pos2]

    return result


def str_find_whitespace(source: str) -> int:
    """
    Localiza e retorna a posição da primeira ocorrência de *whitespace* em *source*.

    Retorna *-1* se nenhum *whitespace* for encontrado.

    :param source: a string a ser pesquisada
    :return: a posição do primeiro whitespace encontrado
    """
    # inicializa a variável de retorno
    result: int = -1

    # busca por whitespace
    for inx, char in enumerate(source):
        if char.isspace():
            result = inx
            break

    return result



from flask import Request
from pathlib import Path
from tempfile import gettempdir
from typing import Final
from werkzeug.datastructures import FileStorage
from .env_pomes import APP_PREFIX, env_get_path

TEMP_DIR: Final[Path] = env_get_path(f"{APP_PREFIX}_TEMP_DIR", Path(gettempdir()))


def file_from_request(request: Request, file_name: str = None, file_seq: int = 0) -> bytes:
    """
    Retrieve and return the contents of the file returned in the response to a request.

    The file may be referred to by its name (*file_name*), or if no name is specified,
    by its sequence number (*file_seq*).

    :param request: the request
    :param file_name: optional name for the file
    :param file_seq:  sequence number for the file, defaults to the first file
    :return: the contents retrieved from the file
    """
    # inicialize the return variable
    result: bytes | None = None

    count: int = len(request.files)
    # has a file been found ?
    if count > 0:
        # yes, retrieve it
        file: FileStorage | None = None
        if isinstance(file_name, str):
            file = request.files.get(file_name)
        elif isinstance(file_seq, int) and file_seq >= 0:
            file_in: str = list(request.files)[file_seq]
            file = request.files[file_in]

        if file:
            result: bytes = file.stream.read()

    return result


def file_get_data(file_data: str | bytes) -> bytes:
    """
    Retrieve and return the data in *file_data* (typeipo *bytes*), or in a file in path *file_data* (tipo *str*).

    :param file_data: file data, or the path to locate the file
    :return: the data
    """
    # declare the return variable
    result: bytes

    # what is the argument type ?
    if isinstance(file_data, bytes):
        # argument is type 'bytes'
        result = file_data

    else:  # isinstance(file_data, str)
        # argumento is a file path
        buf_size: int = 128 * 1024
        file: Path = Path(file_data)
        with file.open("rb") as f:
            file_bytes: bytearray = bytearray()
            while True:
                in_bytes: bytes = f.read(buf_size)
                if in_bytes:
                    file_bytes += in_bytes
                else:
                    break
        result = bytes(file_bytes)

    return result



import inspect
import types


def dict_has_key_chain(source: dict, key_chain: list[str]) -> bool:
    """
    Indicate the existence of an element in *source*, pointed to by the nested key chain *[keys[0]: ... :keys[n]*.

    The path up to he last key in the chain must point to an existing element.
    A given key may indicate the element's position within a *list*, using the format *<key>[<pos>]*.

    :param source: the reference dict
    :param key_chain: the nested key chain
    :return: whether the element exists
    """
    # initialize the return variable
    result: bool = False

    # define the parent el;ement
    parent: dict | None = None

    # does the key chain contain just 1 element ?
    if len(key_chain) == 1:
        # yes, use the provided dict
        parent = source

    # does the key chain contain more than 1 element ?
    elif len(key_chain) > 1:
        # yes, obtain the parent element of the last key in the chain
        parent = dict_get_value(source, key_chain[:-1])

    # is the parent element a dict ?
    if isinstance(parent, dict):
        # yes, proceed
        key = key_chain[-1]

        # is the element denoted by the last key in the chain a list ?
        if key[-1] == "]":
            # yes, recover it
            pos: int = key.find("[")
            inx: int = int(key[pos+1:-1])
            key = key[:pos]
            child = parent.get(key)
            # success, if the element in question is a list with more than 'inx' elements
            result = isinstance(child, list) and len(child) > inx

        # success, if the parent element contains the last key in the chain
        else:
            result = key in parent

    return result


def dict_get_value(source: dict, key_chain: list[str]) -> any:
    """
    Obtain the value of the element in *source*, pointed to by the nested key chain *[keys[0]: ... :keys[n]*.

    The path up to the last key in the chain must point to an existing element.
    A given key may indicate the element's position within a *list*, using the format *<key>[<pos>]*.
    Return *None* if the sought after value is not found.
    Note that returning *None* might not be indicative of the absence of the element in *source*,
    since that element might exist therein with the value *None*. To determine whether this is the case,
    use the operation *dict_has_value*.

    :param source: the reference dict
    :param key_chain: the key chain
    :return: the value obtained
    """
    # initialize the return variable
    result: any = source

    # traverse the keys in the chain
    for key in key_chain:

        # is it possible to proceed ?
        if not isinstance(result, dict):
            # no, terminate the operation
            result = None
            break

        # does the key refer to an elemenent in a list ?
        if key[-1] == "]":
            # yes, retrieve it
            pos: int = key.find("[")
            inx: int = int(key[pos+1:-1])
            result = result.get(key[:pos])

            result = result[inx] if isinstance(result, list) and len(result) > inx else None

            # is it possible to proceed ?
            if isinstance(result, list) and len(result) > inx:
                # yes, proceed
                result = result[inx]
            else:
                # no, abort the operation
                result = None
        else:
            # proceeding is not possible, the element corresponding to 'key' in the dictionary
            result = result.get(key)

    return result


def dict_set_value(target: dict, key_chain: list[str], value: any) -> None:
    """
    Assign to an element of *source* the value *value*.

    The element in question is pointed to by the key chain *[keys[0]: ... :keys[n]*.
    If the element does not exist, it is created with the specified value.
    Any non-existing intermediate elements are created with the value of an empty *dict*.
    A key might indicate the position of the element within a list, using the format *<key>[<pos>]*.
    In such a case, that element must exist.

    :param target: the reference dict
    :param key_chain: the key chain
    :param value: the value to be assigned
    """
    dict_item: any = target
    # traverse the chain, up to end including its penultimate element
    for key in key_chain[:-1]:

        # is it possible to proceed ?
        if not isinstance(dict_item, dict):
            # no, abort the operation
            break

        # does 'key' refer to a list element ?
        if key[-1] == "]":
            # yes, retrieve it
            pos: int = key.find("[")
            inx: int = int(key[pos+1:-1])
            dict_item = dict_item.get(key[:pos])
            # is it possible to proceed ?
            if isinstance(dict_item, list) and len(dict_item) > inx:
                # yes, proceed
                dict_item = dict_item[inx]
            else:
                # no, abort the operation
                dict_item = None
        else:
            # no, does 'dict_item' have 'key' as one of its elements ?
            if key not in dict_item:
                # não, assign to 'dict_item' the element 'key' with an empty dict as value
                dict_item[key] = {}
            dict_item = dict_item.get(key)

    # does a key exist and is 'dict_item'a dict ?
    if len(key_chain) > 0 and isinstance(dict_item, dict):
        # yes, proceed
        key: str = key_chain[-1]
        # does 'key' refer to a list element ?
        if key[-1] == "]":
            # yes, retrieve it
            pos: int = key.find("[")
            inx: int = int(key[pos+1:-1])
            dict_item = dict_item.get(key[:pos])
            # is the assignment possible ?
            if isinstance(dict_item, list) and len(dict_item) > inx:
                # yes, do it
                dict_item[inx] = value
        else:
            # no, assign 'value' to the element 'key' in the dictionary
            dict_item[key] = value


def dict_pop_value(target: dict, key_chain: list[str]) -> any:
    """
    Obtain the value of the element in *source*, pointed to by the nested key chain *[keys[0]: ... :keys[n]*.

    The path up to the last key in the chain must point to an existing element.
    A given key may indicate the element's position within a *list*, using the format *<key>[<pos>]*.
    Return *None* if the sought after value is not found.
    Note that returning *None* might not be indicative of the absence of the element in *source*,
    since that element might exist therein with the value *None*. To determine whether this is the case,
    use the operation *dict_has_value*.

    :param target: the reference dict
    :param key_chain: the key chain
    :return: the value removed
    """
    # initialize the return variable
    result: any = None

    # obtain the element pointed to by the las key in the chain
    parent: dict | None = None

    # does the key chain contain just 1 element ?
    if len(key_chain) == 1:
        # yes, use the provided dict
        parent = target

    # does the key chain contain more than 1 element ?
    elif len(key_chain) > 1:
        # yes, retrieve the parent element of the last key in the chain
        parent = dict_get_value(target, key_chain[:-1])

    # is the parent element a dict ?
    if isinstance(parent, dict):
        # yes, proceed
        key: str = key_chain[-1]

        # does the last key un the chain refer to a list element ?
        if key[-1] == "]":
            # sim, retrieve the list
            pos: int = key.find("[")
            inx: int = int(key[pos+1:-1])
            key = key[:pos]
            child: any = parent.get(key)

            # is the element pointed to by the last key in the chain a list with more than 'inx' elements ?
            if isinstance(child, list) and len(child) > inx:
                # yes, remove that element and return its value
                result = child.pop(inx)

        # does the parent item contain the last key in the chain ?
        elif key in parent:
            # yes, remove that element and return its value
            result = parent.pop(key)

    return result


def dict_replace_value(target: dict, old_value: any, new_value: any) -> None:
    """
    Replace in *target* all occurrences of *old_value* with *new_value*.

    :param target: the reference dict
    :param old_value: the value to be replaced
    :param new_value: the new value
    """
    def list_replace_value(items: list[any], old_val: any, new_val: any) -> None:
        # traverse the list
        for item in items:

            # is 'item' a dict ?
            if isinstance(item, dict):
                # yes, process it recursively
                dict_replace_value(item, old_val, new_val)

            # is 'item' a list ?
            elif isinstance(item, list):
                # yes, process it recursively
                list_replace_value(item, old_val, new_val)

    # traverse the dict
    for curr_key, curr_value in target.items():

        # is 'curr_value' the value to be replaced ?
        if curr_value == old_value:
            # yes, replace it
            target[curr_key] = new_value

        # is 'curr_value' a dict ?
        elif isinstance(curr_value, dict):
            # yes, process it recursively
            dict_replace_value(curr_value, old_value, new_value)

        # is 'curr_value' a list ?
        elif isinstance(curr_value, list):
            # yes, process it recursively
            list_replace_value(curr_value, old_value, new_value)


def dict_get_key(source: dict, value: any) -> any:
    """
    Return the key in *source*, mapping the first occurrence of *value* found.

    Return *None*, if no key is found.
    No recursion is attempted; only the first-level attributes in *source* are inspected.

    :param source: dict to search
    :param value: the reference value
    :return: first key mapping the reference value
    """
    result: any = None
    for key, val in source.items():
        if val == value:
            result = key
            break

    return result


def dict_get_keys(source: dict, value: any) -> list[str]:
    """
    Return all keys in *source*, mapping the value *value*.

    Return *[]* if no key is found.

    :param source: dict to search
    :param value: the reference value
    :return: list containing all keys mapping the reference value
    """
    return [key for key, val in source.items() if val == value]


def dict_merge(target: dict, source: dict) -> None:
    """
    Traverse the elements in *source* to update *target*, according to the criteria presented herein.

    The criteria to be followed are:
      - add the element to *target*, if it does not exist
      - if the element exists in *target*:

        - recursively process both elements, if both are type *dict*
        - add the missing items, if both are type *list*
        - replace the element in *target* se it is a different type, ou if both elements are not the same type

    :param target: the dictionary to be updated
    :param source: the dictionary with the new elements
    """
    # traverse the dictionary with the new elements
    for skey, svalue in source.items():

        # is the item in target ?
        if skey in target:
            # yes, proceed
            tvalue: any = target.get(skey)

            # are both elements dictionaries  ?
            if isinstance(svalue, dict) and isinstance(tvalue, dict):
                # yes, recursively process them
                dict_merge(tvalue, svalue)

            # are both elements lists ?
            elif isinstance(svalue, list) and isinstance(tvalue, list):
                # yes, add the missing elements
                for item in svalue:
                    if item not in tvalue:
                        tvalue.append(item)
            else:
                # both elements are not lists or dictionarie, replace the value in target
                target[skey] = svalue
        else:
            # the item is not in target, add it
            target[skey] = svalue


def dict_coalesce(target: dict, key_chain: list[str]) -> None:
    """
    Coalesce o elemento do tipo *list* em *target* no nível *n*, com a lista no nível imediatamente anterior.

    Esse elemento é apontado pela cadeia de chaves aninhadas *[keys[0]: ... :keys[n]*, e é processado
    como uma sequência de multiplos elementos. Para tanto, as duas últimas chaves da cadeia
    *key_chain* devem estar associadas a valores do tipo *list*.

    :param target: o dicionário a ser coalescido
    :param key_chain: a cadeia de chaves
    """
    # a cadeia de chaves contem mais de 2 chaves ?
    if len(key_chain) > 2:
        # sim, prossiga

        curr_dict: dict | None = target
        # percorre a cadeia até a antepenúltima chave
        for inx, key in enumerate(key_chain[:-2]):

            # é possível prosseguir ?
            if not isinstance(curr_dict, dict):
                # não, aborte a operação
                break

            # key está associado a uma lista ?
            in_list: list[any] = curr_dict.get(key)
            if isinstance(in_list, list):
                # sim, invoque recursivamente o coalescimento dos dicionários da lista
                for in_dict in in_list:
                    # o item da lista é um dicionário ?
                    if isinstance(in_dict, dict):
                        # sim, coalesça-o recursivamente
                        dict_coalesce(in_dict, key_chain[inx + 1:])
                # termina a operação
                curr_dict = None
                break

            # prossiga com o valor associado a key
            curr_dict = curr_dict.get(key)

        # curr_dict é um dicionário contendo a penúltima chave ?
        if isinstance(curr_dict, dict) and \
           isinstance(curr_dict.get(key_chain[-2]), list):
            # sim, prossiga com o coalescimento
            penultimate_elem: list[dict] = curr_dict.pop(key_chain[-2])
            penultimate_list: list[dict] = []

            # percorre os items do penúltimo elemento
            for last_elem in penultimate_elem:

                # last_elem é um dicionário ?
                if isinstance(last_elem, dict):
                    # sim, prossiga
                    outer_dict: dict = {}
                    last_list: list[dict] = []

                    # percorre os itens de last_elem
                    for k1, v1 in last_elem.items():
                        # a chave é a última chave e está associada a uma lista ?
                        if k1 == key_chain[-1] and isinstance(v1, list):
                            # sim, obtenha seus itens para coalescimento posterior
                            for in_dict in v1:
                                # in_dict é um dicionário ?
                                if isinstance(in_dict, dict):
                                    # sim, salve-o coalescido
                                    inner_dict: dict = {}
                                    for k2, v2 in in_dict.items():
                                        inner_dict[k2] = v2
                                    last_list.append(inner_dict)
                                else:
                                    # não, salve-o como está
                                    last_list.append(in_dict)
                        else:
                            # não, coalesça esse item
                            outer_dict[k1] = v1

                    # há itens para coalescimento ?
                    if len(last_list) > 0:
                        # sim, coalesça-os
                        for in_dict in last_list:
                            # in_dict é um dicionário ?
                            if isinstance(in_dict, dict):
                                # sim, acrescente a ele os dados já salvos
                                in_dict.update(outer_dict)
                            # salva o item
                            penultimate_list.append(in_dict)
                    else:
                        # não, salve os itens já coalescidos
                        penultimate_list.append(outer_dict)
                else:
                    # não, salve-o
                    penultimate_list.append(last_elem)

            # substitue a lista original associada à penúltima chave com a nova lista coalescida
            curr_dict[key_chain[-2]] = penultimate_list


def dict_reduce(target: dict, key_chain: list[str]) -> None:
    """
    Realoca os elementos de *target* no nível *n*, para o nível imediatamente acima.

    Esses elementos são apontados pela cadeia de chaves aninhadas *[keys[0]: ... :keys[n]*.
    O elemento no nivel *n* é removido, ao final.

    :param target: o dicionário a ser reduzido
    :param key_chain: a cadeia de chaves
    """
    # a cadeia de chaves contem pelo menos 1 chave ?
    if len(key_chain) > 0:
        # sim, prossiga

        curr_dict: dict | None = target
        # percorre a cadeia até a penúltima chave
        for inx, key in enumerate(key_chain[:-1]):

            # é possível prosseguir ?
            if not isinstance(curr_dict, dict):
                # não, aborte a operação
                break

            # key está associado a uma lista ?
            in_list: list[any] = curr_dict.get(key)
            if isinstance(in_list, list):
                # sim, invoque recursivamente a redução dos dicionários da lista
                for in_dict in in_list:
                    # o item da lista é um dicionário ?
                    if isinstance(in_dict, dict):
                        # sim, reduza-o recursivamente
                        dict_reduce(in_dict, key_chain[inx + 1:])
                # termine a operação
                curr_dict = None
                break

            # prossiga com o valor associado a key
            curr_dict = curr_dict.get(key)

        last_key: str = key_chain[-1]
        # curr_dict contem um dicionário associado a last_key ?
        if isinstance(curr_dict, dict) and \
           isinstance(curr_dict.get(last_key), dict):
            # sim, prossiga com a redução
            last: dict = curr_dict.pop(last_key)
            for key, value in last.items():
                curr_dict[key] = value


def dict_from_list(source: list[dict], key_chain: list[str], value: any) -> dict:
    """
    Localiza em *source*, e retorna, o elemento do tipo *dict* com o valor *value* na cadeia de chaves *key_chain*.

    Retorna *None* se esse *dict* não for encontrado.

    :param source: a lista a ser inspecionada
    :param key_chain: a cadeia de chaves usada na busca
    :param value: o valor associado à cadeia de chaves
    :return: o dict procurado
    """
    # inicializa a variável de retorno
    result: dict | None = None

    for item in source:
        if isinstance(item, dict) and \
           value == dict_get_value(item, key_chain):
            result = item
            break

    return result


def dict_from_object(source: object) -> dict:
    """
    Percorre *source* e cria um *dict* com seus atributos contendo valores não nulos.

    *source* pode ser qualquer objeto, especialmente aqueles que tenham sido decorados com *@dataclass*.

    :param source: o objeto de referência
    :return: dicionário com estrutura equivalente ao objeto de referência
    """
    # inicializa a variável de retorno
    result: dict = {}

    source_module: types.ModuleType = inspect.getmodule(source)
    source_dict: dict = source.__dict__
    for key, value in source_dict.items():
        # value é nulo ou lista vazia ?
        if not (value is None or (isinstance(value, list) and len(value) == 0)):
            # não, prossiga
            name: str = key

            # value é uma lista ?
            if isinstance(value, list):
                # sim, percorra-a
                result[name] = []
                for list_item in value:
                    # list_item é um objeto do mesmo módulo ?
                    if source_module == inspect.getmodule(list_item):
                        # sim, prossiga recursivamente
                        result[name].append(dict_from_object(list_item))
                    else:
                        # não, prossiga linearmente
                        result[name].append(list_item)

            # value é um objeto do mesmo módulo ?
            elif source_module == inspect.getmodule(value):
                # sim, prossiga recursivamente
                result[name] = dict_from_object(value)
            else:
                # não, prossiga linearmente
                result[name] = value

    return result


def dict_transform(source: dict, from_to_keys: list[tuple[str, str]],
                   prefix_from: str = None, prefix_to: str = None) -> dict:
    """
    Constrói um novo *dict*, segundo as regras seguintes.

    Esse dicionário é construído criando-se, para cada elemento da lista de tuplas em
    *from_to_keys*, o elemento indicado pelo segundo termo da tupla, atribuindo-se a ele
    o valor do elemento de *source* indicado pelo primeiro termo da tupla. Ambos os termos
    das tuplas são representados por uma cadeia de chaves aninhadas.

    Os prefixos para as chaves de origem e de destino, se definidos, tem tratamentos distintos.
    São acrescentados na busca de valores em *Source*, e removidos na atribuição de valores
    ao *dict* de retorno.

    :param source: o dict de origem dos valores
    :param from_to_keys: a lista de tuplas contendo as sequências de chaves de origem e destino
    :param prefix_from: prefixo a ser acrescentado às chaves de origem
    :param prefix_to: prefixo a ser removido das chaves de destino
    :return: o novo dicionário
    """
    # import the neeeded functions
    from .list_pomes import list_find_coupled, list_transform, list_unflatten

    # inicializa a variável de retorno
    result: dict = {}

    # percorre o dicionário de origem
    for key, value in source.items():

        # define a cadeia de chaves de origem
        if prefix_from:
            from_keys: str = f"{prefix_from}.{key}"
        else:
            from_keys: str = key

        # obtem a cadeia de chaves de destino
        to_keys: str = list_find_coupled(from_to_keys, from_keys)

        # o destino foi definido ?
        if to_keys:
            # sim, obtenha o valor de destino
            if isinstance(value, dict):
                # valor é um dicionário, transforme-o
                to_value: dict = dict_transform(value, from_to_keys, from_keys, to_keys)
            elif isinstance(value, list):
                # valor é uma lista, transforme-a
                to_value: list = list_transform(value, from_to_keys, from_keys, to_keys)
            else:
                # valor não é dicionário ou lista
                to_value: any = value

            # o prefixo de destino foi definido e ocorre na cadeia de destino ?
            if prefix_to and to_keys.startswith(prefix_to):
                # sim, remova o prefixo
                to_keys = to_keys[len(prefix_to)+1:]
            to_keys_deep: list[str] = list_unflatten(to_keys)

            # atribui o valor transformado ao resultado
            dict_set_value(result, to_keys_deep, to_value)

    return result


def dict_clone(source: dict, from_to_keys: list) -> dict:
    """
    Build a new *dict*, according to the rules presented herein.

    This dictionary is constructed by creating a new element for each element in the list
    *from_to_keys*. When the element of this list is a tuple, the name indicated by its
    second term is used, and the value of the *source* element indicated by the tuple´s
    first term is assigned. This first term can be represented by a chain of  nested keys.
    The name of the element to be created can be omitted, in which case the name of the term
    indicative of the value to be assigned is used. If the corresponding value is not found
    in *source*, *None* is assigned.

    :param source: the source dict
    :param from_to_keys: list of elements indicative of the source and target keys
    :return: the new dict
    """
    # import the neeeded function
    from .list_pomes import list_unflatten

    # inicialize the return variable
    result: dict = {}

    # traverse the list of elements and add to the target dict
    for elem in from_to_keys:
        from_key: str = elem[0] if isinstance(elem, tuple) else elem
        to_key: str = (elem[1] if isinstance(elem, tuple) and len(elem) > 1 else None) or from_key
        result[to_key] = dict_get_value(source, list_unflatten(from_key))

    return result


def dict_listify(target: dict, key_chain: list[str]) -> None:
    """
    Insert the value of the item pointed to by the key chain *[keys[0]: ... :keys[n]* in a list.

    This insertion will happen only if such a value is not itself a list.
    All lists eventually found, up to the penultimate key in the chain, will be processed recursively.

    :param target: the dictionary to be modified
    :param key_chain: the chain of nested keys pointing to the item in question
    """
    def items_listify(in_targets: list, in_keys: list[str]) -> None:

        # traverse the list
        for in_target in in_targets:
            # is the element a dictionary ?
            if isinstance(in_target, dict):
                # yes, process it
                dict_listify(in_target, in_keys)
            # is the element a list ?
            elif isinstance(in_target, list):
                # yes, recursively process it
                # (key chain is also applicable to lists directly nested in lists)
                items_listify(in_target, in_keys)

    parent: any = target
    # traverse the chain up to its penultimate key
    for inx, key in enumerate(key_chain[:-1]):
        parent = parent.get(key)
        # is the item a list ?
        if isinstance(parent, list):
            # yess, process it and close the operation
            items_listify(parent, key_chain[inx+1:])
            parent = None

        # is it possible to proceed ?
        if not isinstance(parent, dict):
            # no, exit the loop
            break

    if isinstance(parent, dict) and len(key_chain) > 0:
        key: str = key_chain[-1]
        # does the item exist and is not a list ?
        if key in parent and not isinstance(parent.get(key), list):
            # yes, insert it in a list
            item: any = parent.pop(key)
            parent[key] = [item]


if __name__ == "__main__":

    s1 = {
        "a0": 0,
        "a1": {
            "b0": "qwert",
            "b1": {
                "c0": None,
                "c1": [1, {"d": [2, {"e": 3}, 4]}, {"d": 5}, {"d": [6, 7]}, [8, 9]]
            }
        }
    }
    mapping = [
        ("a0", "w0"),
        ("a1", "w1"),
        ("a1.b0", "w1.x0"),
        ("a1.b1", "w1.x1"),
        ("a1.b1.c0", "w1.x1.r.y0"),
        ("a1.b1.c1", "w1.x1.y1"),
        ("a1.b1.c1.d", "w1.x1.y1.z")
    ]
    s2 = dict_transform(s1, mapping)

    print(f"original dict:  {s1}")
    keys: list[str] = ["a1", "b1"]
    print(f"reduced chain:  {keys}")
    dict_reduce(s1, keys)
    print(f"reduced dict:   {s1}")
    keys = ["a1", "c1", "d"]
    print(f"listified chain: {keys}")
    dict_listify(s1, keys)
    print(f"listified dict:  {s1}")
    keys = ["a1", "c1", "d"]
    print(f"coalesced chain: {keys}")
    dict_coalesce(s1, keys)
    print(f"coalesced dict:   {s1}")
    print(f"mapping:          {mapping}")
    print(f"transformed dict: {s2}")



import json
import logging
from datetime import datetime
from dateutil import parser
from flask import Request, Response, send_file
from io import BytesIO
from pathlib import Path
from typing import Final, Literal, TextIO
from .datetime_pomes import DATETIME_FORMAT_INV
from .env_pomes import APP_PREFIX, env_get_str, env_get_path
from .file_pomes import TEMP_DIR


def __get_logging_level(level: Literal["debug", "info", "warning", "error", "critical"]) -> int:
    """
    Translate the log severity string *level* into the logging's internal severity value.

    :param level: the string log severity
    :return: the internal logging severity value
    """
    result: int | None
    match level:
        case "debug":
            result = logging.DEBUG          # 10
        case "info":
            result = logging.INFO           # 20
        case "warning":
            result = logging.WARN           # 30
        case "error":
            result = logging.ERROR          # 40
        case "critical":
            result = logging.CRITICAL       # 50
        case _:
            result = logging.NOTSET         # 0

    return result


LOGGING_ID: Final[str] = env_get_str(f"{APP_PREFIX}_LOGGING_ID", f"{APP_PREFIX}")
LOGGING_FORMAT: Final[str] = env_get_str(f"{APP_PREFIX}_LOGGING_FORMAT",
                                         "{asctime} {levelname:1.1} {thread:5d} "
                                         "{module:20.20} {funcName:20.20} {lineno:3d} {message}")
LOGGING_STYLE: Final[str] = env_get_str(f"{APP_PREFIX}_LOGGING_STYLE", "{")

LOGGING_FILE_PATH: Final[Path] = env_get_path(f"{APP_PREFIX}_LOGGING_FILE_PATH",
                                              TEMP_DIR / f"{APP_PREFIX}.log")
LOGGING_FILE_MODE: Final[str] = env_get_str(f"{APP_PREFIX}_LOGGING_FILE_MODE", "a")

# define and configure the logger
PYPOMES_LOGGER: Final[logging.Logger] = logging.getLogger(LOGGING_ID)

# define the logging severity level
# noinspection PyTypeChecker
LOGGING_LEVEL: Final[int] = __get_logging_level(env_get_str(f"{APP_PREFIX}_LOGGING_LEVEL"))

# configure the logger
# noinspection PyTypeChecker
logging.basicConfig(filename=LOGGING_FILE_PATH,
                    filemode=LOGGING_FILE_MODE,
                    format=LOGGING_FORMAT,
                    datefmt=DATETIME_FORMAT_INV,
                    style=LOGGING_STYLE,
                    level=LOGGING_LEVEL)
for _handler in logging.root.handlers:
    _handler.addFilter(logging.Filter(LOGGING_ID))


def logging_get_entries(errors: list[str],
                        log_level: Literal["debug", "info", "warning", "error", "critical"] = None,
                        log_from: str = None, log_to: str = None,
                        log_path: Path | str = LOGGING_FILE_PATH) -> BytesIO:
    """
    Extract and return all entries in the logging file *log_path*.

    It is expected for this logging file to be compliant with *PYPOMES_LOGGER*'s default format.
    The extraction meets the criteria specified by *log_level*, and by the inclusive interval *[log_from, log_to]*.

    :param errors: incidental error messages
    :param log_level: the logging level (defaults to all levels)
    :param log_from: the initial timestamp (defaults to unspecified)
    :param log_to: the finaL timestamp (defaults to unspecified)
    :param log_path: the path of the log file
    :return: the logging entries meeting the specified criteria
    """
    # inicializa variável de retorno
    result: BytesIO | None = None

    # obtain the logging level
    # noinspection PyTypeChecker
    logging_level: int = __get_logging_level(log_level)

    # obtain the initial timestamp
    from_stamp: datetime | None = None
    if log_from:
        from_stamp = parser.parse(log_from)
        if not from_stamp:
            errors.append(f"Value '{from_stamp}' of 'log_from' attribute invalid")

    # obtain the final timestamp
    to_stamp: datetime | None = None
    if log_to:
        to_stamp = parser.parse(log_to)
        if not to_stamp or \
           (from_stamp and from_stamp > to_stamp):
            errors.append(f"Value '{to_stamp}' of 'log_to' attribute invalid")

    file_path: Path = Path(log_path)
    # does the log file exist ?
    if not Path.exists(file_path):
        # no, report the error
        errors.append(f"File '{file_path}' not found")

    # any error ?
    if len(errors) == 0:
        # no, proceed
        result = BytesIO()
        with Path.open(file_path) as f:
            line: str = f.readline()
            while line:
                items: list[str] = line.split(maxsplit=3)
                # noinspection PyTypeChecker
                msg_level: int = __get_logging_level(items[2])
                if msg_level >= logging_level:
                    timestamp: datetime = parser.parse(f"{items[0]} {items[1]}")
                    if (not from_stamp or timestamp >= from_stamp) and \
                       (not to_stamp or timestamp <= to_stamp):
                        result.write(line.encode())
                line = f.readline()

    return result


def logging_get_entries_from_request(request: Request, as_attachment: bool = False) -> Response:
    """
    Retrieve from the log file, and return, the entries matching the criteria specified.

    These criteria are specified in the query string of the HTTP request, according to the pattern
    *path=<log-path>&level=<log-level>&from=YYYYMMDDhhmmss&to=YYYYMMDDhhmmss>*

    All criteria are optional:
        - path: the path of the log file
        - level: the logging level of the entries
        - from: the start timestamp
        - to: the finish timestamp

    :param request: the HTTP request
    :param as_attachment: indicate to browser that it should offer to save the file, or just display it
    :return: file containing the log entries requested on success, or incidental errors on fail
    """
    # declare the return variable
    result: Response

    # initialize the error messages list
    errors: list[str] = []

    # obtain the logging level
    log_level: str = request.args.get("level")

    # obtain the initial and final timestamps
    log_from: str = request.args.get("from")
    log_to: str = request.args.get("to")

    # obtain the path for the log file
    log_path: str = request.args.get("path") or LOGGING_FILE_PATH

    # retrieve the log entries
    # noinspection PyTypeChecker
    log_entries: BytesIO = logging_get_entries(errors, log_level, log_from, log_to, log_path)

    # any error ?
    if len(errors) == 0:
        # no, return the log entries requested as an attached file
        base: str = "entries" if not log_from or not log_to else \
            (
                f"{''.join(ch for ch in log_from if ch.isdigit())}"
                f"{'_'.join(ch for ch in log_to if ch.isdigit())}"
            )
        log_file = f"log_{base}.log"
        log_entries.seek(0)
        result = send_file(path_or_file=log_entries,
                           mimetype="text/plain",
                           as_attachment=as_attachment,
                           download_name=log_file)
    else:
        # yes, report the failure
        result = Response(json.dumps({"errors": errors}), status=401,  mimetype="text/plain")

    return result


def logging_log_msgs(msgs: list[str], output_dev: TextIO = None,
                     log_level: Literal["debug", "info", "warning", "error", "critical"] = "error",
                     logger: logging.Logger = PYPOMES_LOGGER) -> None:
    """
    Write all messages in *msgs* to *logger*'s logging file, and to *output_dev*.

    The output device is tipically *sys.stdout* or *sys.stderr*.

    :param msgs: the messages list
    :param output_dev: output device where the message is to be printed (None for no device printing)
    :param log_level: the logging level, defaults to 'error' (None for no logging)
    :param logger: the logger to use
    """
    # define the log writer
    log_writer: callable = None
    match log_level:
        case "debug":
            log_writer = logger.debug
        case "info":
            log_writer = logger.info
        case "warning":
            log_writer = logger.warning
        case "error":
            log_writer = logger.error
        case "critical":
            log_writer = logger.critical

    # traverse the messages list
    for msg in msgs:
        # has the log writer been defined ?
        if log_writer:
            # yes, log the message
            log_writer(msg)

        # write to output
        __write_to_output(msg, output_dev)


def logging_log_debug(msg: str, output_dev: TextIO = None,
                      logger: logging.Logger = PYPOMES_LOGGER) -> None:
    """
    Write debug-level message *msg* to *logger*'s logging file, and to *output_dev*.

    The output device is tipically *sys.stdout* or *sys.stderr*.

    :param msg: the message to log
    :param output_dev: output device where the message is to be printed (None for no device printing)
    :param logger: the logger to use
    """
    # log the message
    logger.debug(msg)
    __write_to_output(msg, output_dev)


def logging_log_info(msg: str, output_dev: TextIO = None,
                     logger: logging.Logger = PYPOMES_LOGGER) -> None:
    """
    Write info-level message *msg* to *logger*'s logging file, and to *output_dev*.

    The output device is tipically *sys.stdout* or *sys.stderr*.

    :param msg: the message to log
    :param output_dev: output device where the message is to be printed (None for no device printing)
    :param logger: the logger to use
    """
    # log the message
    logger.info(msg)
    __write_to_output(msg, output_dev)


def logging_log_warning(msg: str, output_dev: TextIO = None,
                        logger: logging.Logger = PYPOMES_LOGGER) -> None:
    """
    Write warning-level message *msg* to *logger*'s logging file, and to *output_dev*.

    The output device is tipically *sys.stdout* or *sys.stderr*.

    :param msg: the message to log
    :param output_dev: output device where the message is to be printed (None for no device printing)
    :param logger: the logger to use
    """
    # log the message
    logger.warning(msg)
    __write_to_output(msg, output_dev)


def logging_log_error(msg: str, output_dev: TextIO = None,
                      logger: logging.Logger = PYPOMES_LOGGER) -> None:
    """
    Write error-level message *msg* to *logger*'s logging file, and to *output_dev*.

    The output device is tipically *sys.stdout* or *sys.stderr*.

    :param msg: the message to log
    :param output_dev: output device where the message is to be printed (None for no device printing)
    :param logger: the logger to use
    """
    # log the message
    logger.error(msg)
    __write_to_output(msg, output_dev)


def logging_log_critical(msg: str, output_dev: TextIO = None,
                         logger: logging.Logger = PYPOMES_LOGGER) -> None:
    """
    Write critical-level message *msg* to *logger*'s logging file, and to *output_dev*.

    The output device is tipically *sys.stdout* or *sys.stderr*.

    :param msg: the message to log
    :param output_dev: output device where the message is to be printed (None for no device printing)
    :param logger: the logger to use
    """
    # log the message
    logger.critical(msg)
    __write_to_output(msg, output_dev)


def __write_to_output(msg: str, output_dev: TextIO) -> None:

    # has the output device been defined ?
    if output_dev:
        # yes, write the message to it
        output_dev.write(msg)

        # is the output device 'stderr' ou 'stdout' ?
        if output_dev.name.startswith("<std"):
            # yes, skip to the next line
            output_dev.write("\n")
